""" API Endpoints for Scheduled Tasks
"""
import os
import logging
from flask import jsonify, request
from flask.views import MethodView
from marshmallow import Schema, fields
from flask_smorest import Blueprint
from flask import abort
from marshmallow.exceptions import ValidationError
from pypeline.constants import API_DOC_RESPONSES, API_DOC_PARAMS, API_PATH_V1
from pypeline.utils.config_utils import retrieve_latest_schedule_config
from pypeline.schedule_config_schema import BaseScheduleSchema
from pypeline.flask.decorators import require_accesskey

logger = logging.getLogger(__name__)

bp = Blueprint("schedules", __name__, url_prefix=API_PATH_V1 + "/schedules")


@bp.route("/")
class Schedules(MethodView):
    """Operations related to schedules"""

    @require_accesskey
    @bp.doc(
        responses=API_DOC_RESPONSES,
        parameters=[API_DOC_PARAMS["accesskey"]],
        tags=["Schedules"],
    )
    def get(self):
        """Retrieve list of available schedule entries."""
        access_key = request.headers.get("accesskey")
        try:
            schedule_config = retrieve_latest_schedule_config()
        except ValidationError:
            abort(400, message="Invalid schedule found ...")

        if schedule_config is None:
            abort(404)

        return jsonify(schedule_config)



""" Pipeline APIs
"""
import logging

from flask import jsonify
from flask_smorest import Blueprint
from flask.views import MethodView
from marshmallow import Schema, fields
from marshmallow.exceptions import ValidationError
from webargs.flaskparser import abort

from pypeline.composition import PipelineResult
from pypeline.constants import API_DOC_RESPONSES, API_DOC_PARAMS, API_PATH_V1
from pypeline.utils.pipeline_utils import dag_generator
from pypeline.flask.decorators import require_accesskey
from pypeline.utils.config_utils import retrieve_latest_pipeline_config
from pypeline.pipeline_config_schema import BasePipelineSchema, PipelineSchemaV1

logger = logging.getLogger(__name__)
bp = Blueprint("pipelines", __name__, url_prefix=API_PATH_V1 + "/pipelines")


class InvokePipelineSchema(Schema):
    """Incoming schema for invoking a pipeline"""

    chain_payload = fields.Raw(
        description="Payload contains whatever arguments the pipeline expects "
        "to be passed to each node in the graph.",
        example={"document_id": "123", "send_alert": True},
        required=False,
    )


class InvokePipelineResponseSchema(Schema):
    execution_id = fields.String()
    pipeline_id = fields.String()
    status = fields.String()


class GetPipelineResultResponseSchema(Schema):
    execution_id = fields.String()
    result = fields.Raw()
    result_ttl = fields.Integer()
    results = fields.Raw()
    status = fields.String()
    status_message = fields.String()


@bp.route("/")
class Pipelines(MethodView):
    """Operations against all pipelines."""

    @require_accesskey
    @bp.doc(
        responses=API_DOC_RESPONSES,
        parameters=[API_DOC_PARAMS["accesskey"]],
        tags=["Pipelines"],
    )
    def get(self):
        """Retrieve list of available pipelines."""
        pipeline_config_api_resp = retrieve_latest_pipeline_config()

        if pipeline_config_api_resp is None:
            abort(404)

        try:
            pipelines = []
            for p in pipeline_config_api_resp:
                PipelineSchema = BasePipelineSchema.get_by_version(p["schemaVersion"])
                pipeline_config = PipelineSchema().load(p)
                pipelines.append(pipeline_config)
        except ValidationError as e:
            msg = f"Invalid pipeline configuration: {e}"
            return jsonify({"message": msg}), 202

        return jsonify(pipelines)


@bp.route("/<string:pipeline_id>")
class PipelineInfo(MethodView):
    """Operations against a single pipeline"""

    @require_accesskey
    @bp.doc(
        responses=API_DOC_RESPONSES,
        parameters=[
            API_DOC_PARAMS["accesskey"],
            {
                "in": "path",
                "name": "pipeline_id",
                "description": "pipeline_id for which to retrieve metrics.",
                "type": "string",
                "example": "my_pipeline",
                "required": True,
            },
        ],
        tags=["Pipelines"],
    )
    def get(self, pipeline_id: str):
        """Retrieve details about a specific pipeline."""
        pipeline_config_api_resp = retrieve_latest_pipeline_config(
            pipeline_id=pipeline_id
        )

        if pipeline_config_api_resp is None:
            abort(404)

        try:
            pipeline_config = PipelineSchemaV1().load(pipeline_config_api_resp)
        except ValidationError as e:
            msg = f"Invalid pipeline configuration: {e}"
            return jsonify({"message": msg}), 202

        return jsonify(pipeline_config)


@bp.route("/invoke/<string:pipeline_id>")
class PipelineInvoke(MethodView):
    """Operations involed with pipeline invocation"""

    @require_accesskey
    @bp.doc(
        responses=API_DOC_RESPONSES,
        parameters=[
            API_DOC_PARAMS["accesskey"],
            {
                "in": "path",
                "name": "pipeline_id",
                "description": "pipeline_id for which to retrieve metrics.",
                "type": "string",
                "example": "my_pipeline",
                "required": True,
            },
        ],
        tags=["Pipelines"],
    )
    @bp.arguments(InvokePipelineSchema)
    @bp.response(InvokePipelineResponseSchema)
    def post(self, payload: dict, pipeline_id: str):
        """Invoke a pipeline by it's ID; optionally provide pipeline arguments."""
        pipeline_config = retrieve_latest_pipeline_config(pipeline_id=pipeline_id)

        if pipeline_config is None:
            return abort(404)

        retval = {"pipeline_id": pipeline_id, "status": "starting"}
        try:
            payload = payload["chain_payload"] if "chain_payload" in payload else {}
            pipeline = dag_generator(pipeline_id=pipeline_id, event=payload)
            pipeline.run()
            pipeline_result = PipelineResult(pipeline.execution_id)
            pipeline_result.create_result_entry(pipeline.to_json())
            retval["execution_id"] = pipeline.execution_id
        except Exception as e:
            msg = "Failed to invoke pipeline ... {}".format(pipeline_id)
            logger.error(msg)
            logger.exception(f"{e}")
            abort(500, message=msg)

        return jsonify(retval)


results_responses = API_DOC_RESPONSES.copy()
results_responses[202] = {
    "code": 202,
    "description": "Pipeline is still running. Try again later.",
}
results_responses[204] = {
    "code": 204,
    "description": "The execution results have expired. Re-run pipeline.",
}


@bp.route("/results/<string:execution_id>")
class PipelineResults(MethodView):
    """Operations with respect to pipeline results"""

    @require_accesskey
    @bp.doc(
        responses=results_responses,
        parameters=[
            API_DOC_PARAMS["accesskey"],
            {
                "in": "path",
                "name": "execution_id",
                "description": "execution_id for which to retrieve results",
                "type": "string",
                "example": "4c595cca-9bf1-4150-8c34-6b43faf276c8",
                "required": True,
            },
        ],
        tags=["Pipelines"],
    )
    @bp.response(GetPipelineResultResponseSchema)
    def get(self, execution_id: str):
        """Retrieve results of a pipeline's execution based on execution_id

        NOTE: Cached results expire after a time window so are not available
        forever.
        """
        try:
            pr = PipelineResult(execution_id)
            pr.load()
            retval = {"execution_id": execution_id, "status": pr.status}
            if pr.status == "unavailable":
                retval["status_message"] = "Results expired. Re-run pipeline."
                return retval, 200

            if pr.status == "pending":
                retval["status_message"] = "Results pending. Check again soon."
                return retval, 202

            else:
                retval["status_message"] = "Results available."
                retval["results"] = pr.get_results()
                return retval, 200

        except Exception as e:
            msg = "Failed to retrieve results for execution id: {}".format(execution_id)
            logger.error(msg)
            logger.exception(f"{e}")
            abort(500, message=msg)



""" Initialize most extensions used throughout application
"""
import logging

logger = logging.getLogger(__name__)

try:
    # Client packages *should* provide a `pypeline.yaml` file. This
    # loads the configuration file with the provided name of the client
    # package (e.g. pypeline_demo)
    from pypeline.pypeline_yaml import load_client_config_and_version

    pypeline_config, pypeline_client_version = load_client_config_and_version()
except Exception as e:
    pypeline_config = None
    pypeline_client_version = None
    logger.warning("Unable to load client Pypeline config ... {}".format(e))



import logging
import typing
import networkx as nx
from dramatiq import get_broker, Message
from pypeline.composition import parallel_pipeline
from pypeline.dramatiq import LazyActor, get_callable, register_lazy_actor
from pypeline.utils.config_utils import retrieve_latest_pipeline_config

T = typing.TypeVar("T")  # T can be any type


logger = logging.getLogger(__name__)


def get_execution_graph(
    config: dict,
    adjacency_key: str = "dagAdjacency",
    task_definitions_key: str = "taskDefinitions",
) -> nx.DiGraph:
    """Generate a directed graph based on a pipeline config's adjacency list
    and task definitions.

    `dagAdjacency` is a dictionary containing all nodes and downstream
    nodes.

    `taskDefinitions` is a dictionary containing metadata required for
    each node such as the worker, model version, etc. This metadata is
    attached to each node so it can be retrieved directly from the graph.
    """
    G = nx.DiGraph()

    # Get our adjacency list and task definitions
    adjacency_dict = config.get(adjacency_key, {})
    task_definitions = config.get(task_definitions_key, {})
    if len(adjacency_dict.keys()) == 0:
        logger.warning(
            "Adjacency definition `{}` was not found ...".format(adjacency_key)
        )

    # Build the graph
    for node in adjacency_dict.keys():
        adjacent_nodes = adjacency_dict[node]

        # If no adjacent nodes, then this is a terminal node
        if len(adjacent_nodes) == 0:
            G.add_node(node, attr_dict=task_definitions.get(node, {}))
            continue

        # Otherwise, we'll add an edge from this node to all adjacent nodes
        # and add the task defnition metadata to the edge
        G.add_edges_from(
            [(node, n, task_definitions.get(n, {})) for n in adjacent_nodes]
        )
    return G


def process_non_none_value(value: T) -> None:
    """
    Processes a value that must not be None.

    The function checks if the provided value is None and raises a ValueError if it is.
    If the value is not None, it proceeds to process and print the value.

    :param value: The value to process. Can be of any type, but must not be None.

    :raises ValueError: If the value is None.

    Example:
    >>> process_non_none_value(42)
    Processing value: 42

    >>> process_non_none_value("hello")
    Processing value: hello

    >>> process_non_none_value(None)
    Traceback (most recent call last):
        ...
    ValueError: None value is not allowed
    """
    if value is None:
        raise ValueError("None value is not allowed")


def topological_sort_with_parallelism(
    graph: nx.DiGraph, executable_nodes=None
) -> typing.List[typing.List[T]]:
    """
    Recurse over the graph to find an optimal execution strategy for processing nodes in an order where
    no node shall be processed before all of its predecessors have been processed. The function handles
    parallel execution by identifying nodes that can be processed in parallel at each step. If the graph
    contains a cycle, the function will not be able to generate an execution plan and will raise an exception.

    :param graph: A directed acyclic graph (DiGraph) from NetworkX.
    :param executable_nodes: A list of lists where each inner list contains nodes that can be executed
                              in parallel at each step. This parameter is used for recursion.
    :return: A list of lists where each inner list contains nodes that can be executed in parallel at each step.

    >>> g = nx.DiGraph()
    >>> g.add_edges_from([(1, 2), (1, 3), (2, 4), (3, 4)])
    >>> topological_sort_with_parallelism(g)
    [[1], [2, 3], [4]]

    >>> g = nx.DiGraph()
    >>> g.add_edges_from([(1, 2), (2, 3), (3, 4)])
    >>> topological_sort_with_parallelism(g)
    [[1], [2], [3], [4]]

    >>> g = nx.DiGraph()
    >>> g.add_edges_from([(1, 2), (2, 3), (3, 1)])
    >>> topological_sort_with_parallelism(g)
    Traceback (most recent call last):
        ...
    NetworkXUnfeasible: Graph contains a cycle, cannot compute a topological sort.
    """
    nodes = list(nx.topological_sort(graph))
    round_executable_nodes = [n for n in nodes if graph.in_degree(n) == 0]
    graph.remove_nodes_from(round_executable_nodes)

    if len(round_executable_nodes) == 0:
        return executable_nodes
    else:
        executable_nodes = [] if executable_nodes is None else executable_nodes
        executable_nodes.append(round_executable_nodes)
        return topological_sort_with_parallelism(graph, executable_nodes)


def dag_generator(pipeline_id: str, *args, **kwargs):
    pipeline_config = retrieve_latest_pipeline_config(pipeline_id=pipeline_id)["config"]
    graph = get_execution_graph(pipeline_config)
    optimal_execution_graph = topological_sort_with_parallelism(graph.copy())
    broker = get_broker()

    registered_actors: typing.Dict[str, LazyActor] = {}
    broker.actors.clear()

    messages: typing.List[typing.List[Message]] = []

    task_definitions = pipeline_config["taskDefinitions"]
    for task_group in optimal_execution_graph:
        message_group = []
        for task in task_group:
            module_path = task_definitions[task]["handler"]
            tmp_handler = get_callable(module_path)
            lazy_actor = register_lazy_actor(
                broker, tmp_handler, pipeline_config["metadata"]
            )
            registered_actors[task] = lazy_actor
            if args and not kwargs:
                message_group.append(registered_actors[task].message(*args))
            elif kwargs and not args:
                message_group.append(registered_actors[task].message(**kwargs))
            elif args and kwargs:
                message_group.append(registered_actors[task].message(*args, **kwargs))
            else:
                message_group.append(registered_actors[task].message())
        messages.append(message_group)
    p = parallel_pipeline(messages)

    return p



import importlib
import os.path
import sys
import typing
import logging
import click
from pypeline.extensions import pypeline_config
from warnings import warn
from functools import wraps
from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.triggers.cron import CronTrigger
from typing import Awaitable, Callable, Optional, Union, TYPE_CHECKING, TypeVar
from dramatiq import Broker, Middleware, actor as register_actor, set_broker, get_broker
from dramatiq.brokers.rabbitmq import RabbitmqBroker
from dramatiq.cli import (
    CPUS,
    HAS_WATCHDOG,
    main as dramatiq_worker,
    make_argument_parser as dramatiq_argument_parser,
    import_object,
)
from dramatiq.middleware import default_middleware, CurrentMessage
from dramatiq.results import Results
from dramatiq.results.backends.redis import RedisBackend
from flask import current_app, Flask
from flask.cli import with_appcontext

from pypeline.constants import (
    REDIS_URL,
    RABBIT_URL,
    DEFAULT_BROKER_CALLABLE,
    MS_IN_SECONDS,
    DEFAULT_TASK_TTL,
    DEFAULT_RESULT_TTL,
    DEFAULT_TASK_MAX_RETRY,
    DEFAULT_TASK_MIN_BACKOFF,
    DEFAULT_TASK_MAX_BACKOFF,
)
from pypeline.middleware import ParallelPipeline
from pypeline.utils.config_utils import (
    retrieve_latest_schedule_config,
    get_service_config_for_worker,
)

if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    P = ParamSpec("P")
else:
    P = TypeVar("P")
R = TypeVar("R")

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def configure_default_broker(broker: Broker = None):
    redis_backend = RedisBackend(url=REDIS_URL)
    rabbit_broker = broker if broker is not None else RabbitmqBroker(url=RABBIT_URL)
    rabbit_broker.add_middleware(Results(backend=redis_backend))
    rabbit_broker.add_middleware(ParallelPipeline(redis_url=REDIS_URL))
    rabbit_broker.add_middleware(CurrentMessage())
    register_actors_for_workers(rabbit_broker)
    set_broker(rabbit_broker)


def guess_code_directory(broker):
    actor = next(iter(broker.actors.values()))
    modname, *_ = actor.fn.__module__.partition(".")
    mod = sys.modules[modname]
    return os.path.dirname(mod.__file__)


def get_module(resource_dot_path: str):
    """Retrieve the module based on a 'resource dot path'.
    e.g. package.subdir.feature_file.MyCallable
    """
    module_path = ".".join(resource_dot_path.split(".")[:-1])
    module = importlib.import_module(module_path)
    return module


def get_callable_name(resource_dot_path: str) -> str:
    """Retrieve the callable based on config string.
    e.g. package.subdir.feature_file.MyCallable
    """
    callable_name = resource_dot_path.split(".")[-1]
    return callable_name


def get_callable(resource_dot_path: str) -> Callable:
    """Retrieve the actual handler class based on config string.
    e.g. package.subdir.feature_file.MyCallable
    """
    module = get_module(resource_dot_path)
    callable_name = get_callable_name(resource_dot_path)
    return getattr(module, callable_name)


def register_lazy_actor(
    broker: Broker,
    fn: Optional[Callable[P, Union[Awaitable[R], R]]] = None,
    pipeline_meta: typing.Dict = {},
    **kwargs,
) -> typing.Type["LazyActor"]:
    kwargs["queue_name"] = pipeline_meta.get("queue", "default")
    kwargs["max_retries"] = pipeline_meta.get("maxRetry", DEFAULT_TASK_MAX_RETRY)
    # Convert from seconds to milliseconds
    kwargs["min_backoff"] = (
        pipeline_meta.get("retryBackoff", DEFAULT_TASK_MIN_BACKOFF) * MS_IN_SECONDS
    )
    kwargs["max_backoff"] = (
        pipeline_meta.get("retryBackoffMax", DEFAULT_TASK_MAX_BACKOFF) * MS_IN_SECONDS
    )
    kwargs["time_limit"] = pipeline_meta.get("maxTtl", DEFAULT_TASK_TTL) * MS_IN_SECONDS
    # Always store results for registered pipeline actors
    kwargs["store_results"] = pipeline_meta.get("store_results", False)
    if kwargs["store_results"]:
        kwargs["result_ttl"] = (
            pipeline_meta.get("result_ttl", DEFAULT_RESULT_TTL) * MS_IN_SECONDS
        )
    lazy_actor: LazyActor = LazyActor(fn, kwargs)
    lazy_actor.register(broker)
    return lazy_actor


def register_actors_for_workers(broker: Broker):
    service = get_service_config_for_worker(pypeline_config)
    scheduled_jobs_config = retrieve_latest_schedule_config()

    if not service:
        return
    for task in service.get("registeredTasks", []):
        pipeline_meta = None
        for pipeline_key, pipeline in pypeline_config["pipelines"].items():
            pipeline_config = pipeline["config"]
            pipeline_tasks = [
                t["handler"] for t in pipeline_config["taskDefinitions"].values()
            ]
            if task["handler"] in pipeline_tasks:
                pipeline_meta = pipeline_config["metadata"]
                break

        if pipeline_meta is None:
            for job in scheduled_jobs_config:
                config = job["config"]
                if config["task"] == task["handler"]:
                    pipeline_meta = {"queue": config.get("queue", "default")}

        if pipeline_meta is None:
            raise ValueError(
                f"Registered task {task['handler']} is not defined in a pipeline or scheduled task"
            )

        try:
            worker_path = task["handler"]  # Required, no default
            tmp_handler = get_callable(worker_path)
            if pipeline_meta and pipeline_meta.get("maxRetry", 0) >= 0:
                pipeline_meta["store_results"] = True
                _ = register_lazy_actor(broker, tmp_handler, pipeline_meta)
        except Exception as e:
            print(e)
            print(f"Unable to add a task to dramatiq: {e}")


class Dramatiq:
    """Flask extension bridging Dramatiq broker and Flask app.

    Dramatiq API is eager. Broker initialisation precede actor declaration.
    This breaks application factory pattern and other way to initialize
    configuration after import.

    This class enables lazy initialization of Dramatiq. Actual Dramatiq broker
    is instanciated only once Flask app is created.

    .. automethod:: actor
    .. automethod:: init_app
    """

    def __init__(
        self,
        app: Flask = None,
        name: str = "dramatiq",
        config_prefix: str = None,
        middleware: typing.List[Middleware] = None,
    ):
        """
        :app: Flask application if created. See :meth:`init_app`.

        :param broker_configuration_callable_module: In order to work in fork and spawn mode
            we need to configure our broker using a callable function.  Default is specified as
            "pypeline.flask_dramatiq:configure_default_broker".  This allows the user to
            override if necessary.

        :param name: Unique identifier for multi-broker app.

        :param config_prefix: Flask configuration option prefix for this
            broker. By default, it is derived from ``name`` parameter,
            capitalized.

        :param middleware: List of Dramatiq middleware instances to override
             Dramatiq defaults.

        Flask-Dramatiq always prepend a custom middleware to the middleware
        stack that setup Flask context. This way, every middleware can use
        Flask app context.

        """
        self.actors = []
        self.app = None
        self.config_prefix = config_prefix or name.upper() + "_BROKER"
        self.name = name
        self.broker = None
        if middleware is None:
            middleware = [m() for m in default_middleware]
        self.middleware = middleware
        if app:
            self.init_app(app)

    def __repr__(self) -> str:
        return "<%s %s>" % (self.__class__.__name__, self.name)

    def init_app(self, app: Flask):
        """Initialize extension for one Flask application

        This method triggers Dramatiq broker instantiation and effective actor
        registration.

        """
        if self.app is not None:
            warn(
                "%s is used by more than one flask application. "
                "Actor's context may be set incorrectly." % (self,),
                stacklevel=2,
            )
        self.app = app
        app.extensions["dramatiq-" + self.name] = self

        module_name, broker_or_callable = import_object(DEFAULT_BROKER_CALLABLE)

        # Callable function is expected to setBroker()
        if callable(broker_or_callable):
            print(f"Configuring broker via {DEFAULT_BROKER_CALLABLE}")
            broker_or_callable()
        else:
            raise TypeError("DEFAULT_BROKER_CALLABLE must point to a callable function")
        self.broker = get_broker()
        for actor in self.actors:
            actor.register(broker=self.broker)

    def actor(self, fn=None, **kw):
        """Register a callable as Dramatiq actor.

        This decorator lazily register a callable as a Dramatiq actor. The
        actor can't be called before :meth:`init_app` is called.

        :param kw: Keywords argument passed to :func:`dramatiq.actor`.

        """
        # Substitute dramatiq.actor decorator to return a lazy wrapper. This
        # allows to register actors in extension before the broker is
        # effectively configured by init_app.

        def decorator(fn):
            lazy_actor = LazyActor(self, fn, kw)
            self.actors.append(lazy_actor)
            if self.app:
                lazy_actor.register(self.broker)
            return lazy_actor

        if fn:
            return decorator(fn)
        return decorator


def format_actor(actor):
    return "%s@%s" % (actor.actor_name, actor.queue_name)


def ensure_return_value(default_value=None):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Call the original function
            result = func(*args, **kwargs)
            # Check if the function has returned a value
            if result is None:
                # Return the default value if the function returned None
                return default_value
            return result

        return wrapper

    return decorator


class LazyActor(object):
    # Intermediate object that register actor on broker an call.

    def __init__(self, fn, kw):
        self.fn = fn
        self.kw = kw
        self.actor = None

    def __call__(self, *a, **kw):
        return self.fn(*a, **kw)

    def __repr__(self):
        return "<%s %s.%s>" % (
            self.__class__.__name__,
            self.fn.__module__,
            self.fn.__name__,
        )

    def __getattr__(self, name):
        if not self.actor:
            raise AttributeError(name)
        return getattr(self.actor, name)

    def register(self, broker):
        self.actor = register_actor(
            actor_name=f"{self.fn.__module__}.{self.fn.__name__}",
            broker=broker,
            **self.kw,
        )(ensure_return_value(default_value=True)(self.fn))

    # Next is regular actor API.
    def send(self, *a, **kw):
        return self.actor.send(*a, **kw)

    def message(self, *a, **kw):
        return self.actor.message(*a, **kw)

    def send_with_options(self, *a, **kw):
        return self.actor.send_with_options(*a, **kw)


def list_managed_actors(broker, queues):
    queues = set(queues)
    all_actors = broker.actors.values()
    if not queues:
        return all_actors
    else:
        return [a for a in all_actors if a.queue_name in queues]


@click.command("cron-scheduler")
def cron_scheduler():  # pragma: no cover
    # Configure our broker that we will schedule registered tasks for
    scheduler = BlockingScheduler()
    configure_default_broker()
    broker = get_broker()
    jobs = retrieve_latest_schedule_config()

    for job in jobs:
        if job["enabled"]:
            config = job["config"]
            worker_path = config["task"]
            tmp_handler = get_callable(worker_path)
            pipeline_meta = {"queue": config.get("queue", "default")}
            actor = register_lazy_actor(broker, tmp_handler, pipeline_meta)
            schedule = config["schedule"]
            scheduler.add_job(
                actor.send,
                CronTrigger.from_crontab(
                    f"{schedule['minute']} {schedule['hour']} {schedule['dayOfMonth']} {schedule['monthOfYear']} {schedule['dayOfWeek']}"
                ),
            )

    try:
        scheduler.start()
    except KeyboardInterrupt:
        scheduler.shutdown()


@click.command("pypeline-worker")
@click.argument("broker_name", default="dramatiq")
@click.option(
    "-v", "--verbose", default=0, count=True, help="turn on verbose log output"
)
@click.option(
    "-p",
    "--processes",
    default=CPUS,
    metavar="PROCESSES",
    show_default=True,
    help="the number of worker processes to run",
)
@click.option(
    "-t",
    "--threads",
    default=8,
    metavar="THREADS",
    show_default=True,
    help="the number of worker treads per processes",
)
@click.option(
    "-Q",
    "--queues",
    type=str,
    default=None,
    metavar="QUEUES",
    show_default=True,
    help="listen to a subset of queues, comma separated",
)
@click.option(
    "--use-spawn",
    type=bool,
    default=False,
    metavar="USE_SPAWN",
    show_default=True,
    help="start processes by spawning (default: fork on unix, spawn on windows)",
)
@with_appcontext
def pypeline_worker(
    verbose, processes, threads, queues, broker_name, use_spawn
):  # pragma: no cover
    """Run dramatiq workers.

    Setup Dramatiq with broker and task modules from Flask app.

    \b
    examples:
      # Run dramatiq with 1 thread per process.
      $ flask worker --threads 1

    \b
      # Listen only to the "foo" and "bar" queues.
      $ flask worker -Q foo,bar

    \b
      # Consuming from a specific broker
      $ flask worker mybroker
    """
    # Plugin for flask.commands entrypoint.
    #
    # Wraps dramatiq worker CLI in a Flask command. This is private API of
    # dramatiq.

    parser = dramatiq_argument_parser()

    # Set worker broker globally.
    needle = "dramatiq-" + broker_name
    broker = current_app.extensions[needle].broker
    set_broker(broker)

    command = [
        "--processes",
        str(processes),
        "--threads",
        str(threads),
        # Fall back to flask_dramatiq global broker
        DEFAULT_BROKER_CALLABLE,
    ]

    if use_spawn:
        command += ["--use-spawn"]

    if current_app.config["DEBUG"]:
        verbose = max(1, verbose)
        if HAS_WATCHDOG:
            command += ["--watch", guess_code_directory(broker)]

    queues = queues.split(",") if queues else []
    if queues:
        command += ["--queues"] + queues
    command += verbose * ["-v"]
    args = parser.parse_args(command)
    logger.info("Able to execute the following actors:")
    for actor in list_managed_actors(broker, queues):
        logger.info("    %s.", format_actor(actor))

    dramatiq_worker(args)



import time

import redis


class LockingParallelBarrier:
    def __init__(self, redis_url, task_key="task_counter", lock_key="task_lock"):
        # Connect to Redis using the provided URL
        self.redis = redis.StrictRedis.from_url(redis_url, decode_responses=True)
        self.task_key = task_key
        self.lock_key = lock_key

    def acquire_lock(self, timeout=5):
        """Acquire a lock using Redis."""
        while True:
            if self.redis.set(self.lock_key, "locked", nx=True, ex=timeout):
                return True
            time.sleep(0.1)

    def release_lock(self):
        """Release the lock in Redis."""
        self.redis.delete(self.lock_key)

    def set_task_count(self, count):
        """Initialize the task counter in Redis."""
        self.redis.set(self.task_key, count)

    def decrement_task_count(self):
        """Decrement the task counter in Redis."""
        return self.redis.decr(self.task_key)

    def get_task_count(self):
        """Get the current value of the task counter."""
        return int(self.redis.get(self.task_key) or 0)



""" Pypeline Constants
"""
import os

# Pypeline configuration defaults
PYPELINE_YAML_PATH = os.environ.get("PYPELINE_YAML_PATH", "pypeline.yaml")
PYPELINE_CLIENT_PKG_NAME = os.environ.get("PYPELINE_CLIENT_PKG_NAME", None)
WORKER_NAME = os.environ.get("WORKER_NAME", None)
API_ACCESS_KEY = os.environ.get("API_ACCESS_KEY", None)
DEFAULT_BROKER_CALLABLE = os.environ.get(
    "DEFAULT_BROKER_CLS", "pypeline.dramatiq:configure_default_broker"
)

# Pypeline broker connections
RABBIT_URL = os.environ.get("RABBIT_URL", "amqp://admin:password@127.0.0.1:5672")
REDIS_URL = os.environ.get("REDIS_URL", "redis://localhost:6379/0")

# Pypeline task defaults
PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL = int(
    os.getenv("DRAMATIQ_PARALLEL_PIPELINE_CALLBACK_BARRIER_TTL", "86400000")
)
DEFAULT_RESULT_TTL = int(os.getenv("DEFAULT_RESULT_TTL", 86400))  # seconds (1 day)
DEFAULT_TASK_TTL = int(os.getenv("DEFAULT_TASK_TTL", 600))  # seconds (10 minutes)
DEFAULT_TASK_MAX_RETRY = int(os.getenv("DEFAULT_TASK_MAX_RETRY", 3))
DEFAULT_TASK_MIN_BACKOFF = int(os.getenv("DEFAULT_TASK_MIN_BACKOFF", 15))  # seconds
DEFAULT_TASK_MAX_BACKOFF = int(
    os.getenv("DEFAULT_TASK_MAX_BACKOFF", 3600)
)  # seconds (1 hour)


MS_IN_SECONDS = 1000
API_PATH_V1 = "/api/v1"

# Default 'responses' dictionary when decorating endpoints with @api.doc()
# Extend as necessary.
API_DOC_RESPONSES = {
    200: {"code": 200, "description": "Successful response."},
    400: {"code": 400, "description": "Malformed request. Verify payload is correct."},
    401: {
        "code": 401,
        "description": "Unauthorized. Verify your API Key (`accesskey`) header.",
    },
}

# Default 'params' dictionary when decorating endpoints with @api.doc()
# Extend as necessary.
API_DOC_PARAMS = {
    "accesskey": {
        "in": "header",
        "name": "accesskey",
        "description": "Your API Consumer's `accesskey`",
        "type": "string",
        "required": False,
    }
}

DEFAULT_OPENAPI_CONFIG = (
    ("SWAGGER_UI_DOC_EXPANSION", "list"),
    ("API_DOCUMENTATION_TITLE", "Pypeline API Specs"),
    ("API_DOCUMENTATION_DESCRIPTION", "Available API Endpoints"),
    ("OPENAPI_VERSION", "3.0.2"),
    ("OPENAPI_URL_PREFIX", "/api/v1"),
    ("OPENAPI_SWAGGER_APP_NAME", "Pypeline - API Reference"),
    ("OPENAPI_SWAGGER_UI_PATH", "/docs"),
    ("OPENAPI_SWAGGER_BASE_TEMPLATE", "swagger/swagger_ui.html"),
    ("OPENAPI_SWAGGER_URL", "/docs"),
    (
        "OPENAPI_SWAGGER_UI_URL",
        "https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.24.2/",
    ),
    ("EXPLAIN_TEMPLATE_LOADING", False),
)
